

set -ex



pip check
jupyter server -h
echo "not (delete or merge_config or test_culling or test_restart_kernel[jp_server_config0] or test_connection[jp_server_config0])"
pytest -vvv --cov=jupyter_server -k "not (delete or merge_config or test_culling or test_restart_kernel[jp_server_config0] or test_connection[jp_server_config0])" --cov-report=term-missing:skip-covered --no-cov-on-fail --cov-fail-under=70
exit 0
